/*
 * MIDI player for UNIX (library)
 *
 * 19th, Dec. 1994. by Takanari HAYAMA
 *
 */

#define _MYMIDI_C_
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <varargs.h>
#include <stdlib.h>
#include "midi.h"

#define BUFSZ		4096	/* size of read buffer */
#define NPATTERN	4000	/* max number of patterns */

/* global variables */
extern char *com;
extern int debug;
int newfile;
FILE *midif;

struct midi **eventseq;		/* list of midi event  */
int *eventsz;			/* size of list of midi event */
struct midihead header;		/* info for the song */
int ctrack;			/* current track (track# - 1) */
int cptr;			/* ptr within current track */
int *tflag;

void fdebug(fmt, va_alist)
     char *fmt;
     va_dcl
{
  va_list args;

  if (debug) {
    va_start(args);
    fprintf(stderr, "DEBUG: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
  }
  return;
}

/*
 * initialize global variables
 */

void initmidi()
{
  newfile = 0;
  return;
}

/*
 * open a std. midi file
 */

void openmf(fn)
     char *fn;
{
  midif = stdin;
  if (fn) {
    if ((midif = fopen(fn, "r")) == NULL) {
      fprintf(stderr, "%s: Can't open file %s\n", com, fn);
      exit(-1);
    }
  }
  return;
}

/*
 * close file
 */

void closemf()
{
  fclose(midif);

  return;
}

/*
 * read a byte and return it
 */

int myread()
{
  return(getc(midif));
}

/*
 * fatal message
 */

void fatal(s)
     char *s;
{
  fprintf(stderr, "%s: %s\n", com, s);
  exit(-1);
}

/*
 * write a header
 */

int read_header(format, ntrks, division)
     short format, ntrks, division;
{
  int i, n;

  header.division = (int)division;

  /* check unsupported staff */
  if (format == 2) {
    fprintf(stderr, "%s: Format 2 is not supported yet.\n", com);
    return(-1);
  }
  if (division < 0) {
    fprintf(stderr, "%s: SMPTE is not supported.\n", com);
    return(-1);
  }

  /* ok, here we go */
  header.ntrks = n = (format == 0) ? 1 : (int) ntrks;

  if (debug) {
    fprintf(stderr, "DEBUG: FORMAT=%d\tTRKS=%d(%d)\tDIV=%d\n",
		    format, ntrks, n, division);
  }
  /* allocate memory for event */
  eventseq = (struct midi**)malloc(n * sizeof(struct midi*) + 1);
  eventsz  = (int*)malloc(n * sizeof(int) + 1);
  tflag = (int*)malloc(n * sizeof(int) + 1);
  for(i = 0; i < n; ++i) {
    eventseq[i] = (struct midi*)malloc(sizeof(struct midi) * NPATTERN);
    eventsz[i] = NPATTERN;
    tflag[i] = 0;
  }
  eventsz[n] = 0;
  eventseq[n] = NULL;
  ctrack = cptr = 0;

  return(1);
}

/*
 * clean data
 */

void midi_clean(p, n)
     struct midi *p;
     int n;
{
  int i;

  for(i = 0; i < n; ++i){
    p[i].ctime = p[i].type = p[i].n = 0;
    p[i].data  = NULL;
  }

  return;
}

/*
 * start track.
 */

void midi_starttrack()
{
  if (cptr != 0)
    midi_eot(); /* fail safe */
  midi_clean(eventseq[ctrack], eventsz[ctrack]);
  return;
}

/*
 * move the counter to the next chunk
 */
void midi_inc()
{
  struct midi *dummy;
  int i;

  if (++cptr == eventsz[ctrack]) {
    dummy = (struct midi*)malloc(sizeof(struct midi) * 
				 (eventsz[ctrack] + NPATTERN));
    midi_clean(dummy, eventsz[ctrack] + NPATTERN);
    for(i = 0; i < eventsz[ctrack]; ++i)
      dummy[i] = eventseq[ctrack][i];
    eventsz[ctrack] += NPATTERN;
    free(eventseq[ctrack]);
    eventseq[ctrack] = dummy;
  }

  return;
}

/*
 * make chunk
 */

void make_chunk(ev, n)
     struct midi *ev;
     int n;
{
  ev->type  = 0;
  ev->ctime = Mf_currtime;
  ev->n     = n;
  if (n)
    ev->data  = (unsigned char*)malloc(n + 1);
  return;
}

/*
 * free info
 */

void freeinfo()
{
  int i, j;
  
  for(i = 0; i < header.ntrks; ++i) {
    for(j = 0; j < eventsz[i]; ++j)
      free(eventseq[i][j].data);
    free(eventseq[i]);
  }
  free(eventsz);
  free(eventseq);
  return;
}

/*
 * note on
 */

int midi_note_on(chan, note, vol)
     int chan, note, vol;
{
  make_chunk(&eventseq[ctrack][cptr], 3);

  eventseq[ctrack][cptr].data[0] = MIDI_NOTEON + chan;
  eventseq[ctrack][cptr].data[1] = note;
  eventseq[ctrack][cptr].data[2] = vol;

  midi_inc();
  return(1);
}

/*
 * note off
 */

int midi_note_off(chan, note, vol)
     int chan, note, vol;
{
  make_chunk(&eventseq[ctrack][cptr], 3);

  eventseq[ctrack][cptr].data[0] = MIDI_NOTEOFF + chan;
  eventseq[ctrack][cptr].data[1] = note;
  eventseq[ctrack][cptr].data[2] = vol;

  midi_inc();
  return(1);
}

/*
 * polyphonic pressure (after touch)
 */

int midi_pressure(chan, note, pressure)
     int chan, note, pressure;
{
  fdebug("press.");

  make_chunk(&eventseq[ctrack][cptr], 3);

  eventseq[ctrack][cptr].data[0] = MIDI_PRESSURE + chan;
  eventseq[ctrack][cptr].data[1] = note;
  eventseq[ctrack][cptr].data[2] = pressure;

  midi_inc();
  return(1);
}

/*
 * controller
 */

int midi_control(chan, control, value)
     int chan, control, value;
{
  make_chunk(&eventseq[ctrack][cptr], 3);

  eventseq[ctrack][cptr].data[0] = MIDI_CONTROL + chan;
  eventseq[ctrack][cptr].data[1] = control;
  eventseq[ctrack][cptr].data[2] = value;

  midi_inc();
  return(1);
}

/*
 * program
 */

int midi_program(chan, prog)
     int chan, prog;
{
  fdebug("prog.");

  make_chunk(&eventseq[ctrack][cptr], 2);

  eventseq[ctrack][cptr].data[0] = MIDI_PROGRAM + chan;
  eventseq[ctrack][cptr].data[1] = prog;

  midi_inc();
  return(1);
}

/*
 * chanpressure
 */

int midi_chanpressure(chan, pressure)
     int chan, pressure;
{
  fdebug("chanpres.");

  make_chunk(&eventseq[ctrack][cptr], 2);

  eventseq[ctrack][cptr].data[0] = MIDI_CHANPRES + chan;
  eventseq[ctrack][cptr].data[1] = pressure;

  midi_inc();
  return(1);
}

/*
 * pitchbend
 */

int midi_pitchbend(chan, lsb, msb)
     int chan, lsb, msb;
{
  make_chunk(&eventseq[ctrack][cptr], 3);

  eventseq[ctrack][cptr].data[0] = MIDI_PITCHB + chan;
  eventseq[ctrack][cptr].data[1] = lsb;
  eventseq[ctrack][cptr].data[2] = msb;

  midi_inc();
  return(1);
}


/*
 * sequencer specific meta event
 */

int midi_sqspecific(chan, length, msg)
     int chan, length, msg;
{
  /* we ignore the sequencer specific meta-event */
  fdebug("sqspec.");

  return(1);
}

/*
 * text event
 */

int midi_text(msg_type, length, data)
     int msg_type;
     long length;
     char *data;
{
  int i;

  make_chunk(&eventseq[ctrack][cptr], (length + 2));

  eventseq[ctrack][cptr].type    = 1; /* text event */
  eventseq[ctrack][cptr].data[0] = msg_type;
  for(i = 0; i < length; ++i)
    eventseq[ctrack][cptr].data[i + 1] = data[i];

  eventseq[ctrack][cptr].data[length + 1] = '\0';
  fdebug("text. (%d: %s)", msg_type, &eventseq[ctrack][cptr].data[1]);

  midi_inc();
  return(1);
}

/*
 * tempo
 */

int midi_tempo(tempo)
     int tempo;
{
  unsigned char x[4];
  unsigned int *p;

  fdebug("tempo. (%08x)", tempo);

  p = (unsigned int*)x;
  *p = (unsigned int)tempo;

  make_chunk(&eventseq[ctrack][cptr], sizeof(unsigned long));
  eventseq[ctrack][cptr].type    = 2; /* tempo change (msec/midi-qnote) */
  eventseq[ctrack][cptr].data[0] = x[0];
  eventseq[ctrack][cptr].data[1] = x[1];
  eventseq[ctrack][cptr].data[2] = x[2];
  eventseq[ctrack][cptr].data[3] = x[3];

  midi_inc();
  return(1);
}

/*
 * time sig.
 */

int midi_timesig(num, denom, clocks, qnotes)
     int num, denom, clocks, qnotes;
{
  fdebug("timesig. (ignore)");

  header.num    = num;
  header.denom  = denom;
  header.clocks = clocks;
  header.qnotes = qnotes;
  
  return(1);
}

/*
 * End of Track
 */

int midi_eot()
{
  fdebug("eot.");
  fdebug("ctrack = %d, cptr = %d", ctrack, cptr);

  tflag[ctrack] = 1;
  eventsz[ctrack] = cptr;
  ctrack += 1;
  cptr = 0;
  return(1);
}

/*
 * system exclusive
 */

int midi_sysex(ln, msg)
     int ln;
     unsigned char *msg;
{
  int i;

  fdebug("sysex.");

  if ((msg[0] == 0xf0)&&(msg[ln - 1] == 0xf7)) {
    make_chunk(&eventseq[ctrack][cptr], ln);
    for(i = 0; i < ln; ++i)
      eventseq[ctrack][cptr].data[i] = msg[i];
  }
  else {
    make_chunk(&eventseq[ctrack][cptr], ln + 2);
    for(i = 0; i < ln; ++i)
      eventseq[ctrack][cptr].data[i + 1] = msg[i];
    eventseq[ctrack][cptr].data[0]  = 0xf0;
    eventseq[ctrack][cptr].data[ln] = 0xf7;
  }

  eventseq[ctrack][cptr].type = 3;

  midi_inc();
  return(1);
}

/*
 * for fail safe
 */

int midi_endtrack()
{
  if (tflag[ctrack-1] == 0) {
    --ctrack;
    fdebug("Fail Safe!");
    midi_eot();
  }
  return(1);
}
